#include <map>
#include <list>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <signal.h>

#include <boost/extension/factory.hpp>
#include <boost/extension/shared_library.hpp>
#include <boost/extension/type_map.hpp>
#include <boost/extension/convenience.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>

#include <plugins/plugin.hpp>

#include <ndds/ndds_cpp.h>

#ifndef LIBDIR
#  define LIBDIR ""
#endif

#ifndef SYSCONFDIR
#  define SYSCONFDIR ""
#endif

#define DAEMON_NAME "cavecanem-publisher"

using namespace boost::extensions;
using boost::property_tree::ptree;
using namespace std;


void exit_program(int sig);
DDSDomainParticipant *participant = NULL;

int main()
{
    bool daemonize = true;
    
    setlogmask(LOG_UPTO(LOG_INFO));
    openlog(DAEMON_NAME, LOG_CONS, LOG_USER);
    
    if(daemonize) {
       
		syslog(LOG_INFO, "Starting process");    
        /* Our process ID and Session ID */
        pid_t pid, sid;    
    
        /* Fork off the parent process */
        pid = fork();
        if (pid < 0) {
            exit(EXIT_FAILURE);
        }
        
        /* If we got a good PID, then
           we can exit the parent process. */
        if (pid > 0) {
            exit(EXIT_SUCCESS);
        }
        /* Change the file mode mask */
        umask(0);
        
        /* Create a new SID for the child process */
        sid = setsid();
        if (sid < 0) {
            /* Log the failure */
            exit(EXIT_FAILURE);
        }

        /* Change the current working directory */
        if ((chdir("/")) < 0) {
            /* Log the failure */
            exit(EXIT_FAILURE);
        }
            
        /* Close out the standard file descriptors */
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);      
    }
      
    //Plugins path list and iterator
    list<string> lst;
    list<string>::iterator lst_it;
    //Configuration file
    string sysconf_dir(SYSCONFDIR);
    string lib_dir(LIBDIR);
    string conf_file(sysconf_dir + "cavecanem_publisher.xml");
    string qos_file_path(sysconf_dir + "qos/cavecanem_publisher_qos.xml");
    
    //Default publishing period=1minute
    DDS_Duration_t publishing_period = {60,0};
    int domain_id;
  
    try {
        ptree pt;
        read_xml(conf_file,pt);
        //Reading general information
        publishing_period = {pt.get<int>("general.publishing_period")};
        domain_id = pt.get<int>("general.domain_id");

        //Dealing with plugin definitions
        ptree plugins = pt.get_child("plugins");
        for(ptree::const_iterator it = plugins.begin(); it != plugins.end(); ++it) {
            ptree const& plugin = it->second;
            string lib (plugin.get_child("lib_name").data());
            lib = lib_dir + lib;
            lst.push_front(lib);
        }
    } 
    catch(std::exception &e) {
		syslog(LOG_ERR, "%s",e.what());
		exit(EXIT_FAILURE);
    }
       

    signal(SIGHUP, exit_program);
    signal(SIGTERM, exit_program);
    signal(SIGINT, exit_program);
    signal(SIGQUIT, exit_program);
        
    DDS_ReturnCode_t retcode;
    participant = DDSDomainParticipantFactory::get_instance()->
        create_participant(
		       domain_id, /* Domain ID */
		       DDS_PARTICIPANT_QOS_DEFAULT,/* QoS */
		       NULL,/* Listener */
		       DDS_STATUS_MASK_NONE);
    
    if (participant == NULL) {
		syslog(LOG_ERR, "Unable to create domain participant.");
        exit_program(SIGQUIT);
    }

    
    type_map types;   
    for(lst_it = lst.begin(); lst_it != lst.end(); ++lst_it) {
        if(!load_single_library(types,*lst_it))
            syslog(LOG_ERR, "Error loading plugin");
    }
  
  
    // Retrieve a map of all plugin factories taking an int and indexed
    // by a string from the type_map.
    map<string, factory<plugin, DDSDomainParticipant *> >& factories(types.get());
    if (factories.empty()) {
		syslog(LOG_ERR, "No plugins found!");
        exit_program(SIGQUIT);
    }
  
    map<string,boost::shared_ptr<plugin> > map_plugins;
    for(map<string, factory<plugin, DDSDomainParticipant *> >::iterator it = factories.begin();
        it != factories.end(); ++it) {
        map_plugins[it->first].reset(it->second.create(participant));
    }

    while(true) {
        for(map<string,boost::shared_ptr<plugin> >::iterator it = map_plugins.begin(); it != map_plugins.end() ;++it) 
            map_plugins[it->first]->publish_information();
            
        NDDSUtility::sleep(publishing_period);
    }

	exit(EXIT_SUCCESS);
  
}


void exit_program(int sig)
{
    DDS_ReturnCode_t retcode;
  
    if (participant != NULL)
        retcode = participant->delete_contained_entities();
          
    if (retcode != DDS_RETCODE_OK)
		syslog(LOG_ERR, "Contained entities deletion failed.");
  
    retcode = DDSDomainParticipantFactory::get_instance()->
        delete_participant(participant);
  
    if (retcode != DDS_RETCODE_OK) 
		syslog(LOG_ERR, "Domain Participant Factory Deletion failed.");
  
    syslog(LOG_WARNING, "Received signal %s", strsignal(sig));
    syslog(LOG_INFO, "%s daemon exiting", DAEMON_NAME);
	exit(EXIT_SUCCESS);
}
